describe('ZhLayoutPreferences', function(){

    var zhPreferences;
    var templateManager;
    var eventAggregator;
    var containerMerger;
    var zhUtil;
    var zhPromise;
    var eventEngine;
    var zhLayoutPreferences;
    var messageService;

    beforeEach(function(){
        String.prototype.latinise = function(){
            return this;
        };
        String.prototype.latinize = String.prototype.latinise;

        templateManager = {
            project: {customFloatingActions: []},
            container: {}
        };
        Util.getNestedProperty = function(obj, pathToProperty) {
            var none = {};
            var got = pathToProperty.reduce(function(prev, next) {
                return prev[next] ? prev[next] : none;
            }, obj);
            return got === none ? undefined : got;
        };
        zhPromise = Q;
        zhPreferences   = jasmine.createSpyObj('ZhPreferences',   ['registerRepository', 'getDefaultWidgetPreferences', 'loadPreferences', 'getWidgetPreferences', 'getPreferences', 'savePreferences', 'deletePreferences', 'updateDefaultWidgetPreferences', 'downloadPreferenceFile']);
        eventAggregator = jasmine.createSpyObj('EventAggregator', ['subscribe']);
        containerMerger = jasmine.createSpyObj('ContainerMerger', ['merge', 'restore']);
        zhUtil          = jasmine.createSpyObj('zhUtil',          ['clearRow']);
        eventEngine     = jasmine.createSpyObj('EventEngine',     ['registerEvent']);
        messageService  = jasmine.createSpyObj('messageService',  ['showMessageByCode']);

        zhLayoutPreferences = new ZhLayoutPreferences(zhPreferences, templateManager, eventAggregator, containerMerger, zhPromise, eventEngine, zhUtil, messageService);
    });

    describe('constructor', function(){
        it('should subscribe to beforeProjectStartup event', function(){
            eventAggregator.subscribe.and.callFake(function(eventName, cbk){
                cbk();

                expect(zhPreferences.registerRepository).toHaveBeenCalledWith('zh-layout-preferences#layout');
                expect(templateManager.project.customFloatingActions).toEqual([{
                    'name': 'Layout Action',
                    'label': 'Layout',
                    'type': 'layout',
                    'order': 20,
                    'icon': 'layout',
                    'template': 'zh-layout-preferences#floating-layout-card.html',
                    'expandable': true,
                    'isVisible': jasmine.any(Function)
                }]);
            });

            zhLayoutPreferences = new ZhLayoutPreferences(zhPreferences, templateManager, eventAggregator, containerMerger, zhPromise, eventEngine, zhUtil);

            expect(eventAggregator.subscribe).toHaveBeenCalledWith('beforeProjectStartup', jasmine.any(Function));
        });

        it('should show custom action if widget.floatingControl.customizationAction.layout is not false', function(){
            eventAggregator.subscribe.and.callFake(function(eventName, cbk){
                cbk();
            });

            zhLayoutPreferences = new ZhLayoutPreferences(zhPreferences, templateManager, eventAggregator, containerMerger, zhPromise, eventEngine, zhUtil);

            var widget = {
                'floatingControl': {
                    'searchAction': false,
                    'customizationAction': {
                        'layout': undefined
                    }
                }
            };

            var isVisible = templateManager.project.customFloatingActions[0].isVisible(widget);

            expect(isVisible).toEqual(true);
        });

        it('should hide custom action if widget.floatingControl.customizationAction.layout is false', function(){
            eventAggregator.subscribe.and.callFake(function(eventName, cbk){
                cbk();
            });

            zhLayoutPreferences = new ZhLayoutPreferences(zhPreferences, templateManager, eventAggregator, containerMerger, zhPromise, eventEngine, zhUtil);

            var widget = {
                'floatingControl': {
                    'customizationAction': {
                        'layout': false
                    }
                }
            };

            var isVisible = templateManager.project.customFloatingActions[0].isVisible(widget);

            expect(isVisible).toEqual(false);
        });

        it('should create layout events', function(){
            expect(eventEngine.registerEvent.calls.count()).toEqual(2);
            expect(eventEngine.registerEvent.calls.argsFor(0)).toEqual(['LayoutBeforeApply', 'beforeLayoutApply']);
            expect(eventEngine.registerEvent.calls.argsFor(1)).toEqual(['LayoutAfterApply', 'afterLayoutApply']);
        });
    });

    describe('methods', function(){

        it('updateLayouts', function(done){
            var layouts = [{'ID': 1}, {'ID': 2}];

            spyOn(zhLayoutPreferences, 'getWidgetLayouts').and.callFake(function(){
                return Q.when(layouts);
            });

            var layoutsWithIsDefault = [
                {'ID': 1, 'DEFAULT_LAYOUT': false},
                {'ID': 2, 'DEFAULT_LAYOUT': false}
            ];

            spyOn(zhLayoutPreferences, 'addIsDefaultLayoutToLayouts').and.callFake(function(){
                return Q.when(layoutsWithIsDefault);
            });

            var widgetID = 'widget-1';
            var dataSource = {};

            zhLayoutPreferences.updateLayouts(widgetID, dataSource).then(function(){
                expect(zhLayoutPreferences.getWidgetLayouts).toHaveBeenCalledWith(widgetID);
                expect(zhLayoutPreferences.addIsDefaultLayoutToLayouts).toHaveBeenCalledWith(layouts, widgetID);
                expect(dataSource.data).toEqual(layoutsWithIsDefault);

                zhLayoutPreferences.registerLayoutDataSource(widgetID, dataSource);

                zhLayoutPreferences.updateLayouts(widgetID).then(function(){
                    expect(zhLayoutPreferences.getWidgetLayouts).toHaveBeenCalledWith(widgetID);
                    expect(zhLayoutPreferences.addIsDefaultLayoutToLayouts).toHaveBeenCalledWith(layouts, widgetID);
                    expect(dataSource.data).toEqual(layoutsWithIsDefault);

                    done();
                });
            });
        });

        it('addIsDefaultLayoutToLayouts', function(done){
            zhPreferences.getDefaultWidgetPreferences.and.callFake(function(){
                return Q.when({
                    'DEFAULT_LAYOUT': 2
                });
            });

            var layouts = [{'ID': 1}, {'ID': 2}, {'ID': 3}];
            var widgetID = 'widget-1';

            zhLayoutPreferences.addIsDefaultLayoutToLayouts(layouts, widgetID).then(function(result){
                expect(zhPreferences.getDefaultWidgetPreferences).toHaveBeenCalledWith(widgetID);
                expect(result).toEqual([
                    {'ID': 1, 'DEFAULT_LAYOUT': false},
                    {'ID': 2, 'DEFAULT_LAYOUT': true},
                    {'ID': 3, 'DEFAULT_LAYOUT': false}
                ]);

                done();
            });
        });

        it('getWidgetLayouts', function(done){
            zhPreferences.getWidgetPreferences.and.callFake(function(){
                return Q.when('getWidgetLayouts result');
            });

            var widgetID = 'widget-1';
            zhLayoutPreferences.getWidgetLayouts(widgetID).then(function(result){
                expect(zhPreferences.getWidgetPreferences).toHaveBeenCalledWith('layout', widgetID);
                expect(result).toEqual('getWidgetLayouts result');
                done();
            });
        });

        it('getLayouts', function(done){
            zhPreferences.getPreferences.and.callFake(function(){
                return Q.when('getPreferences result');
            });

            var query = 'query example';
            zhLayoutPreferences.getLayouts(query).then(function(result){
                expect(zhPreferences.getPreferences).toHaveBeenCalledWith('layout', query);
                expect(result).toEqual('getPreferences result');
                done();
            });
        });

        it('saveLayouts', function(done){
            zhUtil.clearRow.and.callFake(function(row){
                return row;
            });

            zhPreferences.savePreferences.and.callFake(function(){
                return Q.when('savePreferences result');
            });

            var layouts = [{'ID': 1}, {'ID': 2}];
            zhLayoutPreferences.saveLayouts(layouts).then(function(result){
                expect(zhPreferences.savePreferences).toHaveBeenCalledWith('layout', layouts);
                expect(result).toEqual('savePreferences result');
                done();
            });
        });

        it('deleteLayouts', function(done){
            zhUtil.clearRow.and.callFake(function(row){
                return row;
            });

            zhPreferences.deletePreferences.and.callFake(function(){
                return Q.when('deletePreferences result');
            });

            var layouts = [{'ID': 1}, {'ID': 2}];
            zhLayoutPreferences.deleteLayouts(layouts).then(function(result){
                expect(zhPreferences.deletePreferences).toHaveBeenCalledWith('layout', layouts);
                expect(result).toEqual('deletePreferences result');
                done();
            });
        });

        it('applyOriginalLayout', function(done){
            var widget = {
                'redraw': jasmine.createSpy('redraw'),
                'setAppliedLayout': jasmine.createSpy('setAppliedLayout')
            };

            zhLayoutPreferences.applyOriginalLayout(widget).then(function(){
                expect(containerMerger.restore).toHaveBeenCalledWith(widget, 'ORIGINAL');
                expect(widget.redraw).toHaveBeenCalled();
                expect(widget.setAppliedLayout).toHaveBeenCalledWith({'ID': 'ORIGINAL', 'USER_ID': 0, 'ORGANIZATION_ID': 0, 'LAYOUT_CHANGES': '{}', 'LABEL_TO_PREFERENCE': 'Original'});

                done();
            });
        });

        it('applyOriginalLayout: with events', function(done){
            var widget = {
                'beforeLayoutApply': jasmine.createSpy('beforeLayoutApply'),
                'afterLayoutApply': jasmine.createSpy('afterLayoutApply'),
                'redraw': jasmine.createSpy('redraw'),
                'setAppliedLayout': jasmine.createSpy('setAppliedLayout')
            };

            zhLayoutPreferences.applyOriginalLayout(widget).then(function(){
                expect(containerMerger.restore).toHaveBeenCalledWith(widget, 'ORIGINAL');
                expect(widget.redraw).toHaveBeenCalled();
                expect(widget.setAppliedLayout).toHaveBeenCalledWith({'ID': 'ORIGINAL', 'USER_ID': 0, 'ORGANIZATION_ID': 0, 'LAYOUT_CHANGES': '{}', 'LABEL_TO_PREFERENCE': 'Original'});

                expect(widget.beforeLayoutApply).toHaveBeenCalledWith({'layout': {'ID': 'ORIGINAL', 'USER_ID': 0, 'ORGANIZATION_ID': 0, 'LAYOUT_CHANGES': '{}', 'LABEL_TO_PREFERENCE': 'Original'}});
                expect(widget.afterLayoutApply).toHaveBeenCalledWith({'layout': {'ID': 'ORIGINAL', 'USER_ID': 0, 'ORGANIZATION_ID': 0, 'LAYOUT_CHANGES': '{}', 'LABEL_TO_PREFERENCE': 'Original'}});

                done();
            });
        });

        it('applyLayout', function(done){
            var widget = {
                'redraw': jasmine.createSpy('redraw'),
                'setAppliedLayout': jasmine.createSpy('setAppliedLayout')
            };

            var layout = {
                'ID': 1,
                'LAYOUT_CHANGES': '{"fields": [{"name": "test"}]}',
                'PDF_REPORT_ORIENTATION': 'LANDSCAPE'
            };

            Util.extend = function(){};
            var container = new Container();
            container.id = 1;
            container.name =  'Container';

            templateManager.container = container;

            zhLayoutPreferences.applyLayout(widget, layout).then(function(){
                expect(containerMerger.merge).toHaveBeenCalledWith(container, {"fields": [{"name": "test"}]}, 'ORIGINAL');
                expect(widget.redraw).toHaveBeenCalled();
                expect(widget.setAppliedLayout).toHaveBeenCalledWith({'ID': 1, 'LAYOUT_CHANGES': '{"fields": [{"name": "test"}]}', 'PDF_REPORT_ORIENTATION': 'LANDSCAPE'});
                expect(widget.reportOrientation).toEqual('LANDSCAPE');

                done();
            });
        });

        it('applyLayout: should update widget backupPDFOrientation on first call', function(done){
            var widget = {
                'redraw': function(){},
                'setAppliedLayout': function(){},
                'reportOrientation': 'LANDSCAPE'
            };

            var layout = {
                'ID': 1,
                'LAYOUT_CHANGES': '{"fields": [{"name": "test"}]}'
            };

            zhLayoutPreferences.applyLayout(widget, layout).then(function(){
                expect(widget.backupPDFOrientation).toEqual('LANDSCAPE');
            }).then(function(){
                widget.reportOrientation = 'PORTRAIT';
                return zhLayoutPreferences.applyLayout(widget, layout);
            }).then(function(){
                expect(widget.backupPDFOrientation).toEqual('LANDSCAPE');
                done();
            });
        });

        it("applyLayout: shouldn't add deleted fields", function(done){
            var widget = new Widget();
            widget.redraw = function(){};
            widget.setAppliedLayout = function(){};
            widget.reportOrientation = 'LANDSCAPE';
            widget.id = '1';
            widget.name = 'widgetTest';
            widget.fields = [
                {
                    id: '11'
                },
                {
                    id: '12'
                },
                {
                    id: '13'
                },
                {
                    id: '14'
                },
                {
                    id: '15'
                }
            ];

            var changes = {
                widgets: [{
                    id: '1',
                    fields: [
                        {
                            id: '31'
                        },
                        {
                            id: '15'
                        },
                        {
                            id: '12'
                        },
                        {
                            id: '32'
                        },
                        {
                            id: '11'
                        }
                    ]
                }]
            };

            var layout = {
                'ID': 1,
                'LAYOUT_CHANGES': JSON.stringify(changes),
                'PDF_REPORT_ORIENTATION': 'LANDSCAPE'
            };

            Util.extend = function(){};
            var container = new Container();
            container.id = 1;
            container.name =  'Container';
            container.widgets = [widget];

            templateManager.container = container;

            zhLayoutPreferences.applyLayout(widget, layout).then(function(){
                var preparedChanges = {
                    widgets: [{
                        id: '1',
                        fields: [
                            {
                                id: '15'
                            },
                            {
                                id: '12'
                            },
                            {
                                id: '11'
                            }
                        ]
                    }]
                };
                expect(containerMerger.merge).toHaveBeenCalledWith(container, preparedChanges, 'ORIGINAL');
                done();
            });
        });

        it("applyLayout: shouldn't apply undefined property from container", function(done){
            var widget = new Widget();
            widget.redraw = function(){};
            widget.setAppliedLayout = function(){};
            widget.reportOrientation = 'LANDSCAPE';
            widget.id = '1';
            widget.name = 'widgetTest';
            widget.fields = [
                {
                    id: '11'
                },
                {
                    id: '12'
                },
                {
                    id: '13'
                },
                {
                    id: '14'
                },
                {
                    id: '15'
                }
            ];

            var changes = {
                widgets: [{
                    id: '1',
                    fields: [
                        {
                            id: '31'
                        },
                        {
                            id: '15'
                        },
                        {
                            id: '12'
                        },
                        {
                            id: '32'
                        },
                        {
                            id: '11'
                        }
                    ]
                }]
            };

            var layout = {
                'ID': 1,
                'LAYOUT_CHANGES': JSON.stringify(changes),
                'PDF_REPORT_ORIENTATION': 'LANDSCAPE'
            };

            Util.extend = function(){};
            var container = new Container();
            container.id = 1;
            container.name =  'Container';
            container.widgets = undefined;

            templateManager.container = container;

            zhLayoutPreferences.applyLayout(widget, layout).then(function(){
                var preparedChanges = {
                    widgets: []
                };
                expect(containerMerger.merge).toHaveBeenCalledWith(container, preparedChanges, 'ORIGINAL');
                done();
            });
        });

        it('registerLayoutDataSource', function(){
            var widget = {
                'id': '1'
            };

            var dataSource = {
                'data': []
            };

            var result = zhLayoutPreferences.registerLayoutDataSource(widget.id, dataSource);

            expect(result).toEqual({'1': dataSource});

            result = zhLayoutPreferences.registerLayoutDataSource(undefined, dataSource);

            expect(result).toEqual({'1': dataSource});
        });

        it('downloadPreferenceFile', function(done){
            var layout = {
                'ID': 'ORIGINAL',
                'USER_ID': 'user-1',
                'WIDGET_ID': '1',
                'LAYOUT_CHANGES': '{}',
                'LABEL_TO_PREFERENCE': 'Original Layout',
                'PDF_REPORT_ORIENTATION': 'portrait',
                'LAST_MODIFIED_DATE': '01/01/2017'
            };

            var layoutToExport = {
                'ID': 'ORIGINAL',
                'LABEL_TO_PREFERENCE': 'Original Layout',
                'WIDGET_ID': '1',
                'PDF_REPORT_ORIENTATION': 'portrait',
                'LAYOUT_CHANGES': {},
                'LAST_MODIFIED_DATE': '01/01/2017'
            };

            zhPreferences.downloadPreferenceFile.and.callFake(function(layout, fileName){
                return Q.when({'layout': layout, 'fileName': fileName});
            });

            zhLayoutPreferences.downloadPreferenceFile(layout).then(function(){
                var _layout = arguments[0].layout;
                var _fileName = arguments[0].fileName;
                expect(_layout).toEqual(layoutToExport);
                expect(_fileName).toBe('originalLayout.layout');

                done();
            });
        });

        it('validateLayoutImport', function(){
            var listWidget = {
                'id': '1',
                'fields': [{
                    'name': 'field_1'
                },{
                    'name': 'field_2'
                }]
            };

            var layoutFile = {
                'ID': 999,
                'WIDGET_ID': '2',
                'LABEL_TO_PREFERENCE': 'Layout'
            };

            var importLayoutWidget = {
                'currentRow': {
                    'LABEL_TO_PREFERENCE': 'layout name'
                },
                'isValid': function(){
                    return true;
                },
                'configWidget': {
                    'listWidget': listWidget,
                    'dataSource': {
                        'data': [{
                            'ID': 999,
                            'LABEL_TO_PREFERENCE': 'Layout'
                        }]
                    }
                }
            };

            var messageCallCount = 0;
            messageService.showMessageByCode.and.callFake(function(){
                messageCallCount++;
            });

            expect(zhLayoutPreferences.validateLayoutImport(layoutFile, importLayoutWidget.configWidget)).toBe(false); // wrong widget id
            layoutFile.WIDGET_ID = '1';

            expect(zhLayoutPreferences.validateLayoutImport(layoutFile, importLayoutWidget.configWidget)).toBe(false); // duplicated layout id
            layoutFile.ID = 998;

            expect(zhLayoutPreferences.validateLayoutImport(layoutFile, importLayoutWidget.configWidget)).toBe(false); // duplicated layout name
            layoutFile.LABEL_TO_PREFERENCE = 'new layout';

            expect(zhLayoutPreferences.validateLayoutImport(layoutFile, importLayoutWidget.configWidget)).toBe(true);
            expect(messageCallCount).toEqual(3);
        });
    });
});